// simple CPPM decoder for Arduino (Uno / Nano / Pro Mini / etc. with either a 16MHz or 8MHz crystal.
// ceptimus.  June 2021
// connections (seem like a weird order, but these were inherited from the 'stand alone, no crystal, ATmega328 version')
// Port ref | Arduino label name   | Function |
//   D2     |         2            | CPPM in  |
//   B1     |         9            | Servo 1  |
//   B2     |        10            | Servo 2  |
//   B3     |        11            | Servo 3  |
//   B4     |        12            | Servo 4  |
//   B5     |        13            | Servo 5  |
//   C0     |        A0            | Servo 6  |
//   C1     |        A1            | Servo 7  |
//   C2     |        A2            | Servo 8  |

// constant for determining minimum sync pulse width depends on the clock speed.  This constant works out to be 40 for a 16 MHz crystal and 20 for an 8 MHz crystal
#define TWO_POINT_FIVE_MILLISECOND_COUNT (F_CPU / 400000)

void setup() {
  noInterrupts(); // disable all interrupts for the entire sketch
  PORTB = 0b00000000; // first five servo outputs off
  DDRB  = 0b00111110; // five output pins for the first five servos
  PORTC = 0b00000000; // last three servo outputs off
  DDRC  = 0b00000111; // three output pins for the last three servos 
  TCCR0A = 0x00;  // free-running timer/counter 0, counting up to 0xFF and wrapping back to zero
  TCCR0B = 0x05;  // with an 8 MHz clock, the counter will increment every 128 us. With a 16 MHz clock it will increment every 64 us.  Anything over TWO_POINT_FIVE_MILLISECOND_COUNT is a sync pulse
  uint8_t previousCount = TCNT0;  // initialize previousCount by getting current count from timer/counter 0
  while (true) { // loop forever
    asm( // inline assembly for fastest possible speed in 'waiting for edges' loops: sbrX takes up to 3 cycles, and rjmp takes 2 cycles: loops ~0.31 microseconds with 16MHz clock
      "loopA: sbic %0, %1 \n" // skip next instruction if bit 2 of PIND is clear
      "rjmp loopA \n" // otherwise keep testing the bit - this waits for low state on CPPM input pin
      "loopB: sbis %0, %1 \n" // skip next instruction if bit 2 of PIND is set
      "rjmp loopB \n" // otherwise keep testing the bit - this waits for high state on CPPM input pin
      : : "I" (_SFR_IO_ADDR(PIND)), "I" (PORTD2)
    );
    uint8_t currentCount = TCNT0;  // take current count from timer/counter 0
    if ((uint8_t)(currentCount - previousCount) > TWO_POINT_FIVE_MILLISECOND_COUNT) { 
      PORTB = 0b00000010; // sync pulse detected - so switch first servo output on, and all others off
      PORTC = 0b00000000;
    } else {
      PORTB += PORTB; // switch the next servo output on (and the previous one off)
      PORTC += PORTC;
      if (PORTB == 0b01000000) PORTC = 0b00000001; // after fifth servo on PORTB, activate first on PORT C
    }
    previousCount = currentCount; // the current count from this pass through the loop becomes the previous count for the next pass
  }
}

void loop() {}  // loop() not used because we have an endless loop inside setup();
